"""Fedinesia - deletes old statuses from fediverse accounts. This tool was previously
called MastodonAmnesia
Copyright (C) 2021, 2022, 2023  Mark S Burgunder.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import asyncio
import json
import sys
from datetime import datetime
from math import ceil
from pathlib import Path

import msgspec
import msgspec.json
import msgspec.toml
from httpx import AsyncClient
from loguru import logger as log
from minimal_activitypub import Status
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError
from minimal_activitypub.client_2_server import NetworkError
from minimal_activitypub.client_2_server import NotFoundError
from minimal_activitypub.client_2_server import RatelimitError
from stamina import retry_context
from tqdm import tqdm
from tqdm import trange
from whenever import Instant

from fedinesia import PROGRESS_ID_KEY
from fedinesia.config import Configuration
from fedinesia.util import AuditLog
from fedinesia.util import should_keep


class FediHelper(msgspec.Struct):
    """Class to contain logic for collecting statuses to be deleted."""

    audit_log: AuditLog
    config: Configuration
    oldest_to_keep: datetime
    instance: ActivityPub | None = None
    statuses: list[Status] = msgspec.field(default_factory=list)
    user_info: dict[str, str] | None = None

    async def connect(self) -> None:
        """Connect to Fedi instance."""
        client = AsyncClient(http2=True, timeout=30)

        self.instance = ActivityPub(
            instance=self.config.mastodon.instance,
            access_token=self.config.mastodon.access_token,
            client=client,
        )
        await self.instance.determine_instance_type()
        self.user_info = await self.instance.verify_credentials()

    async def retrieve_and_filter_statuses(
        self,
        continue_max_id: str | None = None,
        limit: int | None = None,
    ) -> None:
        """Retrieve all statuses for account."""
        max_id = continue_max_id
        pagination_progress = ""

        if not self.instance or not self.user_info:
            return

        log.opt(colors=True).info(
            f"We are removing statuses older than <cyan>{self.oldest_to_keep}</> "
            f"from <cyan>{self.config.mastodon.instance}@{self.user_info['username']}</> "
            f"with {self.user_info['statuses_count']} statuses"
        )

        while True:
            pagination_max_id = self.instance.pagination["next"]["max_id"]
            if pagination_max_id and pagination_progress and (pagination_max_id > pagination_progress):
                log.debug("Break")
                break

            max_id = self.determine_pagination(
                pagination_progress=pagination_progress,
                continue_max_id=continue_max_id,
            )

            try:
                log.debug(f"{max_id=}")
                statuses = await self.instance.get_account_statuses(
                    account_id=self.user_info["id"],
                    max_id=max_id,
                )
            except RatelimitError:
                await self.sleep_off_ratelimiting()

            log.debug(f"scrolling - {len(statuses)=}")
            if len(statuses) == 0:
                break

            log.debug(f"oldest_to_keep={self.oldest_to_keep.isoformat()}")
            self.statuses.extend(self._filter_statuses(statuses=statuses))
            log.debug(f"{len(self.statuses)=}")

            if limit and len(self.statuses) > limit:
                statuses_to_delete = self.statuses[:limit]
                self.statuses = statuses_to_delete
                log.debug(f"{len(self.statuses)=}")
                break

            last_status = statuses[-1]
            pagination_progress = last_status.get("id", "")
            log.debug(f"{pagination_progress=}")

    def determine_pagination(self, pagination_progress: str, continue_max_id: str | None) -> str | None:
        """Determine max_id for pagination while taking into account any potential saved progress."""
        if not self.instance or not self.user_info:
            log.error(f"Empty instance ({self.instance=}) or no user_info ({self.user_info=})")
            return None

        max_id = continue_max_id

        log.debug(f"{max_id=}")
        log.debug(f"{pagination_progress=}")
        log.debug(f"{self.instance.pagination=}")

        pagination_max_id = self.instance.pagination["next"]["max_id"]
        if not pagination_max_id:
            pagination_max_id = max_id
            log.debug(f"{pagination_max_id=}")

        if (not max_id) and pagination_progress:
            max_id = pagination_progress
            log.debug(f"{max_id=}")

        if max_id and pagination_max_id and max_id > pagination_max_id:
            max_id = pagination_max_id
            log.debug(f"{max_id=}")

        return max_id

    async def delete_statuses(
        self,
        save_progress_path: Path | None,
    ) -> None:
        """Delete all statuses that should be deleted."""
        log.info(
            f"A record of all deleted statuses will be recorded in the audit log file at {self.audit_log.audit_log}"
        )
        self.audit_log.begin()

        title = "Deleting statuses"
        total_statuses_to_delete = len(self.statuses)
        for status in tqdm(
            iterable=self.statuses,
            desc=f"{title:.<60}",
            ncols=120,
            total=total_statuses_to_delete,
            unit="statuses",
            position=0,
            bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} at {rate_fmt}",
        ):
            try:
                for attempt in retry_context(on=NetworkError):
                    with attempt:
                        await self.delete_single_status(status=status)

                if save_progress_path:
                    last_deleted_id = status if isinstance(status, str) else status["id"]
                    with save_progress_path.open(mode="wb") as save_progress_file:
                        save_progress_file.write(msgspec.toml.encode({PROGRESS_ID_KEY: last_deleted_id}))
            except RatelimitError:
                await self.sleep_off_ratelimiting()

        self.audit_log.end()

        log.info(f"All old statuses deleted! Total of {len(self.statuses)} statuses deleted")

    async def delete_single_status(self, status: Status) -> None:
        """Delete  single status."""
        if not self.instance or not self:
            return

        log.debug(f"delete_single_status(status={status['id']}, instance={self.instance})")
        try:
            await self.instance.delete_status(status=status)
            log.debug(f"delete_single_status - Deleted status {status.get('url')} from {status.get('created_at')}")
            self.audit_log.add_entry(status=status)
        except NotFoundError:
            log.debug("Status for deletion not found. No problem then :)")
        except ActivityPubError as error:
            log.debug(f"delete_single_status - encountered error: {error}")
            log.debug(f"delete_single_status - status: {json.dumps(status, indent=4)}")
            raise error

    def delete_dry_run(self) -> None:
        """Print out what statuses would be deleted for a dry run."""
        log.opt(colors=True).info("--dry-run or -d specified. <yellow><bold>No statuses will be deleted</></>")
        for status in self.statuses:
            log.opt(colors=True).info(
                f"<red>Would</red> delete status {status.get('url')} from {status.get('created_at')}"
            )
        log.opt(colors=True).info(f"<bold>Total of {len(self.statuses)} statuses would be deleted.</>")

    def _filter_statuses(self, statuses: list[Status]) -> list[Status]:
        """Filter out statuses that should not be deleted."""
        deletion_list: list[Status] = []
        for status in statuses:
            if should_keep(status=status, oldest_to_keep=self.oldest_to_keep, config=self.config):
                continue

            deletion_list.append(status)

        log.debug(f"{len(deletion_list)=}")
        return deletion_list

    async def sleep_off_ratelimiting(self) -> None:
        """Wait for rate limiting to be over."""
        if not self.instance:
            log.error("Fedi instance not set, can't continue!")
            sys.exit(-1)

        log.debug(
            f"sleep_off_ratelimiting - Rate limited: Limit: {self.instance.ratelimit_remaining} - "
            f"resetting at: {self.instance.ratelimit_reset}"
        )
        now = Instant.now().py_datetime()
        need_to_wait = ceil((self.instance.ratelimit_reset - now).total_seconds())

        bar_title = f"Waiting until {self.instance.ratelimit_reset:%H:%M:%S %Z} to let server 'cool-down'"
        for _i in trange(
            need_to_wait,
            desc=f"{bar_title:.<60}",
            unit="s",
            ncols=120,
            bar_format="{desc}: {percentage:3.0f}%|{bar}| Eta: {remaining} - Elapsed: {elapsed}",
            position=1,
        ):
            await asyncio.sleep(1)



"""Helper classes and methods to assist with the general function of this bot.

Fedinesia - deletes old statuses for a fediverse account (Mastodon or Pleroma and forks)
Copyright (C) 2021, 2022, 2023  Mark S Burgunder

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
from pathlib import Path
from typing import TypeVar

import msgspec
import msgspec.json
import msgspec.toml
from httpx import AsyncClient
from minimal_activitypub import Visibility
from minimal_activitypub.client_2_server import ActivityPub
from minimal_activitypub.client_2_server import ActivityPubError

from fedinesia import CLIENT_WEBSITE
from fedinesia import USER_AGENT

BC = TypeVar("BC", bound="BotConfig")
ConfigClass = TypeVar("ConfigClass", bound="Configuration")
MC = TypeVar("MC", bound="MastodonConfig")


class BotConfig(msgspec.Struct):
    """Class holding configuration values for general behaviour of Fedinesia."""

    delete_after: int
    skip_deleting_pinned: bool = False
    skip_deleting_faved: bool = False
    skip_deleting_bookmarked: bool = False
    skip_deleting_poll: bool = False
    skip_deleting_visibility: list[Visibility] = []
    skip_deleting_media: bool = False
    skip_deleting_faved_at_least: int = 0
    skip_deleting_boost_at_least: int = 0
    skip_deleting_reactions_at_least: int = 0

    def get_config_values(self: BC) -> None:
        """Initialise instance."""
        self._get_delete_after()
        self._get_skip_bookmarked()
        self._get_skip_faved()
        self._get_skip_pinned()
        self._get_skip_poll()
        self._get_skip_dm()
        self._get_skip_media()
        self._get_skip_faved_at_least()
        self._get_skip_boost_at_lease()
        self._get_skip_reactions_at_lease()

    def _get_skip_poll(self: BC) -> None:
        """Private method to get skip deleting polls value from user if this
        value has not yet been configured.
        """
        print("Should polls be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_poll = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_poll = True
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_dm(self: BC) -> None:
        """Private method to get skip deleting 'private' messages value from
        user if this value has not yet been configured.
        """
        print("Should Direct Messages be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_visibility = []
        elif y_or_n in ("N", "n"):
            self.skip_deleting_visibility = [Visibility.DIRECT]
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_media(self: BC) -> None:
        """Private method to get skip deleting statuses with media value from
        user if this value has not yet been configured.
        """
        print("Should Statuses with attachments / pictures be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_media = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_media = True
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved_at_least(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        favourited value from user if this value has not yet been
        configured.
        """
        print(
            "Should statuses being favourited a certain minimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        print("(enter 0 to disregard this setting)")
        self.skip_deleting_faved_at_least = int(input("[..] Please enter number: "))

    def _get_skip_boost_at_lease(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        boosted value from user if this value has not yet been configured.
        """
        print(
            "Should statuses being boosted a certain minimum number of times be "
            "excluded from deletion even when they get old enough?"
        )
        print("(enter 0 to disregard this setting)")
        self.skip_deleting_boost_at_least = int(input("[..] Please enter number: "))

    def _get_skip_reactions_at_lease(self: BC) -> None:
        """Private method to get skip deleting statuses that have been
        boosted value from user if this value has not yet been configured.
        """
        print(
            "Should statuses with a certain minimum number of reactions be "
            "excluded from deletion even when they get old enough?"
        )
        print("(enter 0 to disregard this setting)")
        self.skip_deleting_reactions_at_least = int(input("[..] Please enter number: "))

    def _get_skip_pinned(self: BC) -> None:
        """Private method to get skip deleting pinned statuses value from
        user if this value has not yet been configured.
        """
        print("Should pinned statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_pinned = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_pinned = True
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_faved(self: BC) -> None:
        """Private method to get skip deleting favorited statuses value from
        user if this value has not yet been configured.
        """
        print("Should favoured statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_faved = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_faved = True
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_skip_bookmarked(self: BC) -> None:
        """Private method to get skip deleting bookmarked statuses from user
        if this value has not yet been configured.
        """
        print("Should bookmarked statuses be deleted when they get old enough?")
        y_or_n = input("[..] Please enter Y for yes or N for no: ")
        if y_or_n in ("Y", "y"):
            self.skip_deleting_bookmarked = False
        elif y_or_n in ("N", "n"):
            self.skip_deleting_bookmarked = True
        else:
            print("! ERROR ... please only respond with 'Y' or 'N'")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

    def _get_delete_after(self: BC) -> None:
        """Private method to get delete after value from user if this value
        has not yet been configured.
        """
        print('Please enter maximum age of retained statuses in the format of "number unit"')
        print('For example "1 weeks" or "3 days". Supported units are:')
        print(" - seconds\n - minutes\n - hours\n - days\n - weeks\n - months")
        max_age = input("[..] Minimum age to delete statuses (in seconds): ")
        max_age_parts = max_age.split(" ")
        max_age_number = int(max_age_parts[0])
        max_age_unit = max_age_parts[1]
        if max_age_unit == "seconds":
            self.delete_after = max_age_number
        elif max_age_unit == "minutes":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "hours":
            self.delete_after = max_age_number * 3600
        elif max_age_unit == "days":
            self.delete_after = max_age_number * 3600 * 24
        elif max_age_unit == "weeks":
            self.delete_after = max_age_number * 3600 * 24 * 7
        elif max_age_unit == "months":
            self.delete_after = max_age_number * 3600 * 24 * 30
        else:
            print("! Error ... unknown unit ({max_age_unit}) specified")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)


class MastodonConfig(msgspec.Struct):
    """Class holding configuration values for Mastodon settings."""

    instance: str
    access_token: str

    @classmethod
    async def establish_new_mastodon_config(cls: type[MC]) -> MC:
        """Establish Mastodon configuration from scratch."""
        instance = input("[..] Enter instance (domain name) for Mastodon account host: ")

        try:
            async with AsyncClient(http2=True) as client:
                # Create app
                client_id, client_secret = await ActivityPub.create_app(
                    instance_url=instance,
                    client=client,
                    user_agent=USER_AGENT,
                    client_website=CLIENT_WEBSITE,
                )

                # Get Authorization Code / URL
                authorization_request_url = await ActivityPub.generate_authorization_url(
                    instance_url=instance,
                    client_id=client_id,
                    user_agent=USER_AGENT,
                )
                print(f"Please go to the following URL and follow the instructions:\n{authorization_request_url}")
                authorization_code = input("[...] Please enter the authorization code:")

                # Validate authorization code and get access token
                access_token = await ActivityPub.validate_authorization_code(
                    client=client,
                    instance_url=instance,
                    authorization_code=authorization_code,
                    client_id=client_id,
                    client_secret=client_secret,
                )

        except ActivityPubError as error:
            print(f"! Error when setting up Fediverse connection: {error}")
            print("! Cannot continue. Exiting now.")
            sys.exit(1)

        return cls(instance=instance, access_token=access_token)


class Configuration(msgspec.Struct):
    """Dataclass to hold all settings for fedinesia."""

    bot: BotConfig
    mastodon: MastodonConfig


async def setup_shop(
    config_file: str,
) -> Configuration:
    """Process command line arguments, establish debug logging to file if
    specified, load configuration.

    :returns:
        Configuration: config for this run of the Fedinesia.
    """
    config_file_path = Path(config_file)
    if config_file_path.exists():
        with config_file_path.open(mode="rb") as config_file_file:
            config: Configuration = msgspec.json.decode(config_file_file.read(), type=Configuration)

    else:
        bot = BotConfig(delete_after=30)
        bot.get_config_values()
        mastodon = await MastodonConfig.establish_new_mastodon_config()
        config = Configuration(bot=bot, mastodon=mastodon)

        config_encoded = msgspec.json.encode(config)
        config_json = msgspec.json.format(config_encoded, indent=4)
        with config_file_path.open(mode="wb") as config_file_file:
            config_file_file.write(config_json)

    return config



"""Fedinesia - deletes old statuses for a fediverse account (Mastodon or Pleroma and forks)
Copyright (C) 2021, 2022, 2023  Mark S Burgunder.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import csv
import json
import logging
import os
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any
from typing import TypeVar

from loguru import logger as log
from minimal_activitypub.client_2_server import Status
from whenever import Instant

from fedinesia import __display_name__
from fedinesia.config import Configuration

logger = logging.getLogger(__display_name__)


AL = TypeVar("AL", bound="AuditLog")


@dataclass
class AuditLog:
    """Class to control the creation of an audit log and adding entries to
    it.
    """

    class Style(str, Enum):
        """Enumerating different audit log file styles implemented."""

        PLAIN = "PLAIN"
        CSV = "CSV"

    audit_log: Path | None
    style: Style = Style.PLAIN
    _audit_log_file: Any | None = None

    def begin(self: AL) -> None:
        """Open audit log file for writing and appending new log entries."""
        if not self.audit_log:
            return

        self._audit_log_file = self.audit_log.open(mode="at")

        if self.style == AuditLog.Style.CSV and os.fstat(self._audit_log_file.fileno()).st_size == 0:
            self._add_csv_header()

    def end(self: AL) -> None:
        """Close audit log file."""
        if self._audit_log_file:
            self._audit_log_file.close()

    def add_entry(self: AL, status: Status) -> None:
        """Append an entry/status details to the audit log file."""
        if self.style == AuditLog.Style.PLAIN:
            self._add_plain_entry(status)
        elif self.style == AuditLog.Style.CSV:
            self._add_csv_entry(status)

    def _add_plain_entry(self: AL, status: Status) -> None:
        """Append a plain text entry/status details to the audit log file."""
        if not self._audit_log_file:
            return

        entry = (
            f"{datetime.now():%Y-%m-%d %H:%M:%S%z} -"
            f" Removed {'poll' if status.get('poll', False) else 'status'} {status.get('url')}"
            f" created @ {status.get('created_at')}"
            f" with {status.get('visibility')} visibility, {len(status.get('media_attachments', []))} attachments."
            f" This status was reblogged {status.get('reblogs_count')} times and"
            f" favourited {status.get('favourites_count')} times."
            f" The status was {'pinned' if status.get('pinned') else 'not pinned'}.\n"
        )
        self._audit_log_file.write(entry)

    def _add_csv_entry(self: AL, status: Status) -> None:
        """Append a csv formatted status details to the audit log file."""
        if not self._audit_log_file:
            return

        csv_writer = csv.writer(self._audit_log_file, quoting=csv.QUOTE_ALL)
        record = [
            datetime.now().strftime("%Y-%m-%d %H:%M:%S%z"),
            "poll" if status.get("poll", False) else "status",
            status.get("url"),
            Instant.parse_iso(status.get("created_at", "0")).py_datetime(),
            status.get("visibility"),
            len(status.get("media_attachments", [])),
            status.get("reblogs_count"),
            status.get("favourites_count"),
            status.get("pinned"),
        ]
        csv_writer.writerow(record)

    def _add_csv_header(self: AL) -> None:
        """Write CSV header line to the audit log file.

        This method is normally called when an audit log file is being
        initially created before writing any other entries to the audit
        log; i.e. The record it writes will be the header record for the
        CSV file.
        """
        if not self._audit_log_file:
            return

        csv_writer = csv.writer(self._audit_log_file, quoting=csv.QUOTE_ALL)
        record = [
            "date and time of deletion",
            "type of status,",
            "url of status before deletion",
            "date and time deleted status was created",
            "visibility",
            "# of media_attachments",
            "reblogs_count",
            "favourites_count",
            "pinned",
        ]
        csv_writer.writerow(record)


def _is_status_too_recent(status: Status, oldest_to_keep: datetime) -> bool:
    """Check if status is too young to delete."""
    status_created_at = Instant.parse_iso(status.get("created_at", "0")).py_datetime()
    if status_created_at >= oldest_to_keep:
        log.debug(f"{status.get('id')} posted at {status.get('created_at')}")
        return True

    return False


def _is_self_bookmarked(status: Status, config: Configuration) -> bool:
    """Check if the status has been bookmarked by the original poster and bookmarked statuses should be kept."""
    if config.bot.skip_deleting_bookmarked and bool(status.get("bookmarked")):
        return True

    return False


def _is_self_favourite(status: Status, config: Configuration) -> bool:
    """Check if the status has been favoured by the original poster and favoured statuses should be kept."""
    if config.bot.skip_deleting_faved and bool(status.get("favourited")):
        return True

    return False


def _is_pinned(status: Status, config: Configuration) -> bool:
    """Check if the status has been pinned and pinned statuses should be kept."""
    if config.bot.skip_deleting_pinned and bool(status.get("pinned")):
        return True

    return False


def _is_poll(status: Status, config: Configuration) -> bool:
    """Check if the status is a poll and if polls should be kept."""
    if config.bot.skip_deleting_poll and bool(status.get("poll")):
        return True

    return False


def _matches_visibility(status: Status, config: Configuration) -> bool:
    """Check if the status visibility matches any visibility that should be kept."""
    if status.get("visibility") in config.bot.skip_deleting_visibility:
        return True

    return False


def _has_media(status: Status, config: Configuration) -> bool:
    """Check if status includes media and if statuses with media should be kept."""
    medias = status.get("media_attachments")
    if config.bot.skip_deleting_media and isinstance(medias, list) and bool(len(medias)):
        return True

    return False


def _has_enough_favourites(status: Status, config: Configuration) -> bool:
    """Check if status has been favoured more often than defined for keeping favoured statuses."""
    favourites = status.get("favourites_count", 0)
    if config.bot.skip_deleting_faved_at_least and favourites >= config.bot.skip_deleting_faved_at_least:
        return True

    return False


def _has_enough_reblogs(status: Status, config: Configuration) -> bool:
    """Check if status has been reblogged more often than defined for keeping reblogged statuses."""
    reblogs = status.get("reblogs_count", 0)
    if config.bot.skip_deleting_boost_at_least and reblogs >= config.bot.skip_deleting_boost_at_least:
        return True

    return False


def should_keep(status: Status, oldest_to_keep: datetime, config: Configuration) -> bool:
    """Determine if status should be kept."""
    keep_rules: list[bool] = [
        _is_status_too_recent(status=status, oldest_to_keep=oldest_to_keep),
        _is_self_bookmarked(status=status, config=config),
        _is_self_favourite(status=status, config=config),
        _is_pinned(status=status, config=config),
        _is_poll(status=status, config=config),
        _matches_visibility(status=status, config=config),
        _has_media(status=status, config=config),
        _has_enough_favourites(status=status, config=config),
        _has_enough_reblogs(status=status, config=config),
    ]

    if any(keep_rules):
        return True

    if config.bot.skip_deleting_reactions_at_least:
        reactions_threshold = config.bot.skip_deleting_reactions_at_least
    else:
        reactions_threshold = 0

    return enough_reactions_to_keep(
        status=status,
        reactions_threshold=reactions_threshold,
    )


def enough_reactions_to_keep(status: Status, reactions_threshold: int) -> bool:
    """Check if emoji_reactions are reported in status and check of they are at least at the
    threshold for keeping the status.
    Note: Reactions are a Pleroma specific feature.

    :param status: Status being examined for reactions
    :param reactions_threshold: Minimum number of reactions to consider the check to be met

    :returns:
        True if the total number of reactions to status is equal to or more than the reactions_threshold.
        False in any other case, including if the status does not contain any reaction counts at all.
    """
    if not (pleroma := status.get("pleroma")):
        return False

    if not (reactions := pleroma.get("emoji_reactions")):
        return False

    log.debug(f"enough_reactions_to_keep - {reactions_threshold=} - reactions={json.dumps(reactions, indent=4)}")

    total_reactions = 0
    for reaction in reactions:
        if emo_count := reaction.get("count"):
            total_reactions += emo_count

    return total_reactions >= reactions_threshold



"""Fedinesia - deletes old statuses from fediverse accounts. This tool was previously
called MastodonAmnesia
Copyright (C) 2021, 2022, 2023  Mark S Burgunder.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import asyncio
import sys
from pathlib import Path
from typing import Annotated

import msgspec
import msgspec.json
import msgspec.toml
import stamina
import typer
from loguru import logger as log
from whenever import Instant

from fedinesia import PROGRESS_ID_KEY
from fedinesia import __display_name__
from fedinesia import __version__
from fedinesia.config import setup_shop
from fedinesia.fedi import FediHelper
from fedinesia.util import AuditLog

stamina.instrumentation.set_on_retry_hooks([])


@log.catch
async def main(  # noqa:PLR0913
    config_file: str,
    is_dry_run: bool,
    audit_log_file: Path | None,
    audit_log_style: AuditLog.Style | None,
    limit: int | None,
    save_progress_path: Path | None,
    continue_from_saved_progress: bool,
) -> None:
    """Perform app function."""
    config = await setup_shop(config_file=config_file)

    oldest_to_keep = Instant.now().subtract(seconds=config.bot.delete_after).py_datetime()

    log.info(f"Welcome to {__display_name__} {__version__}")
    log.debug(f"{config=}")

    continue_max_id: str | None = None
    if continue_from_saved_progress and save_progress_path and save_progress_path.exists():
        with save_progress_path.open(mode="rb") as progress_file:
            saved_progress = msgspec.toml.decode(progress_file.read())
            continue_max_id = saved_progress.get(PROGRESS_ID_KEY)

    if audit_log_style is None:
        audit_log_style = AuditLog.Style.PLAIN

    fedi_helper = FediHelper(
        config=config,
        oldest_to_keep=oldest_to_keep,
        audit_log=AuditLog(audit_log=audit_log_file, style=audit_log_style),
    )
    await fedi_helper.connect()

    await fedi_helper.retrieve_and_filter_statuses(continue_max_id=continue_max_id, limit=limit)

    # If dry-run has been specified, print out list of statuses that would be deleted
    if is_dry_run:
        fedi_helper.delete_dry_run()

    # Dry-run has not been specified... delete statuses!
    else:
        await fedi_helper.delete_statuses(save_progress_path=save_progress_path)


def start() -> None:
    """Start app."""
    typer.run(typer_async_shim)


def typer_async_shim(  # noqa: PLR0913
    config_file: Annotated[str, typer.Option("-c", "--config-file")] = "config.json",
    audit_log_file: Annotated[
        Path | None,
        typer.Option("-a", "--audit-log-file", file_okay=True, dir_okay=False, writable=True),
    ] = None,
    audit_log_style: AuditLog.Style | None = None,
    limit: Annotated[int | None, typer.Option("-l", "--limit")] = None,
    dry_run: Annotated[bool, typer.Option("-d", "--dry-run/--no-dry-run")] = False,
    save_progress: Annotated[
        Path | None,
        typer.Option("-s", "--save-progress", file_okay=True, dir_okay=False, writable=True),
    ] = None,
    continue_from_saved_progress: Annotated[bool, typer.Option("--continue/--no-continue")] = False,
    logging_config_path: Annotated[
        Path | None,
        typer.Option(
            "--logging-config",
            help="Full Path to logging config file",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ] = None,
) -> None:
    """Delete fediverse history. For more information look at https://codeberg.org/MarvinsMastodonTools/fedinesia."""
    if logging_config_path:
        with logging_config_path.open(mode="rb") as log_config_file:
            logging_config = msgspec.toml.decode(log_config_file.read())

        for handler in logging_config.get("handlers"):
            if handler.get("sink") == "sys.stdout":
                handler["sink"] = sys.stdout

        log.configure(**logging_config)

    if continue_from_saved_progress and (not save_progress):
        raise typer.BadParameter("--continue can only be specified if --save-progress has also been specified.")

    if audit_log_style and (not audit_log_file):
        raise typer.BadParameter("--audit-log-style can only be specified if --audit-log-file has also been specified.")

    if audit_log_file and (not audit_log_style):
        audit_log_style = AuditLog.Style.PLAIN

    asyncio.run(
        main(
            config_file=config_file,
            is_dry_run=dry_run,
            audit_log_file=audit_log_file,
            audit_log_style=audit_log_style,
            limit=limit,
            save_progress_path=save_progress,
            continue_from_saved_progress=continue_from_saved_progress,
        )
    )



"""Package 'fedinesia' level definitions."""

from importlib.metadata import version
from typing import Final

__version__: Final[str] = version(str(__package__))

__package_name__: Final[str] = str(__package__)
__display_name__: Final[str] = str(__package__.title()) if __package__ else __package_name__
USER_AGENT: Final[str] = f"{__display_name__}"

CLIENT_WEBSITE: Final[str] = "https://codeberg.org/marvinsmastodontools/fedinesia"

PROGRESS_ID_KEY = "last_deleted_id"
